import { Format } from "./song";
import { GeneralMidiInstruments, noteByMidi } from "../midi/types";
import { createPlayHeadClock, toSignature, createTempo } from "../metronome";
import { patchPlayer } from "./audio";
import { AndyPatch, AndyPercussionPatch } from './patchBank';
const MAX_CHANNELS = 16;
export const NoOp = () => { };
export function createRun(system, song) {
    if (song.format === Format.MultiAsync)
        throw new Error("Currently only support synchronous tracks");
    const run = {
        song: song,
        tracks: [],
        output: system.context().createGain()
    };
    for (let i = 0; i < song.numberOfTracks; i++) {
        const pan = system.context().createPanner();
        pan.orientationX.setValueAtTime(0, 0);
        pan.orientationY.setValueAtTime(0, 0);
        pan.orientationZ.setValueAtTime(-1, 0);
        pan.positionX.setValueAtTime(0, 0);
        pan.positionY.setValueAtTime(1, 0);
        pan.positionZ.setValueAtTime(0, 0);
        const vol = system.context().createGain();
        pan.connect(vol);
        vol.connect(run.output);
        vol.gain.setValueAtTime(1, 0);
        const startFns = new Array(MAX_CHANNELS);
        for (let i = 0; i < MAX_CHANNELS; i++) {
            startFns[i] = [];
        }
        const t = {
            lastFn: startFns,
            idx: 0,
            vol,
            pan,
        };
        run.tracks.push(t);
    }
    run.output.connect(system.masterGainNode());
    return run;
}
export function destroyRun(system, run) {
    run.output.disconnect(system.masterGainNode());
}
const useInstrument = (system, run, patchNum) => {
    switch (patchNum) {
        case 34: {
            const p = AndyPatch.get(GeneralMidiInstruments.ChurchOrgan);
            if (p)
                return patchPlayer(system, p);
        }
        case 31:
        case 30:
        case 29:
        case 26: {
            const p = AndyPatch.get(GeneralMidiInstruments.AcousticBass);
            if (p)
                return patchPlayer(system, p);
        }
        default: {
            const patch = AndyPatch.get(GeneralMidiInstruments.ChurchOrgan);
            if (patch)
                return patchPlayer(system, patch);
        }
    }
    return NoOp;
};
const usePercussion = (system, run, patchNum) => {
    switch (patchNum) {
        default: {
            const patch = AndyPercussionPatch.get(patchNum);
            if (patch)
                return patchPlayer(system, patch);
        }
    }
    return NoOp;
};
export function silenceRun(songRun) {
    songRun.tracks.forEach(track => {
        for (let i = 0; i < MAX_CHANNELS; i++) {
            const fns = track.lastFn[i];
            while (fns.length) {
                const soundOff = fns.pop();
                if (soundOff)
                    soundOff();
            }
        }
    });
}
export function rewindRun(songRun) {
    songRun.tracks.forEach(track => {
        track.idx = 0;
    });
}
export function playRun(system, run) {
    let tempo = createTempo(run.song.ticksPerQuarterNote, 500000);
    let sig = toSignature("4/4");
    const tickCallBack = (tick, time, bi) => {
        var _a, _b, _c;
        const trackLen = run.song.tracks.length;
        const stopAllSounds = (track) => {
            for (let i = 0; i < MAX_CHANNELS; i++) {
                const fns = track.lastFn[i];
                while (fns.length) {
                    const soundOff = fns.pop();
                    if (soundOff)
                        soundOff();
                }
            }
        };
        for (let t = 0; t < trackLen; t++) {
            const track = run.tracks[t];
            const events = run.song.tracks[t].events;
            for (let e = track.idx; e < events.length; e++) {
                const ce = events[e];
                if (ce.abstime !== tick) {
                    track.idx = e;
                    break;
                }
                if (ce.type === "Meta") {
                    if (ce.subtype === "SetTempo") {
                        tempo = createTempo(tempo.ppq, ce.data[0]);
                    }
                    if (ce.subtype === "TimeSignature") {
                        sig.noteBeats = ce.data[0];
                        sig.perMeasure = ce.data[1];
                        sig.ticksInMetronome = ce.data[2];
                        sig.toQuarter32nd = ce.data[3];
                    }
                }
                if (ce.type === "Channel") {
                    if (ce.subtype === "ProgramChange") {
                        track.instrument = useInstrument(system, run, ce.data[0]);
                    }
                    if (ce.subtype === "Controller") {
                        if (ce.data[0] === 0x7) {
                            track.vol.gain.setValueAtTime((ce.data[1] / 127), time);
                        }
                        if (ce.data[0] === 0xa) {
                            const amt = (ce.data[1] - 127 + 63.5) / 63.5;
                            const amtScaled = Math.round(amt * 10) / 10;
                            track.pan.positionX.setValueAtTime(amtScaled, time);
                        }
                    }
                    if (ce.subtype === "NoteOff") {
                        const fns = track.lastFn[(_a = ce.channel) !== null && _a !== void 0 ? _a : 0];
                        while (fns.length) {
                            const soundOff = fns.pop();
                            if (soundOff)
                                soundOff();
                        }
                    }
                    if (ce.subtype === "NoteOn") {
                        const mnote = (_b = ce.data[0]) !== null && _b !== void 0 ? _b : 60;
                        const note = noteByMidi(mnote);
                        if (ce.channel === 0x9) {
                            track.lastFn[ce.channel].push(usePercussion(system, run, mnote)(note, system.currentTime(), track.pan));
                        }
                        else {
                            if (track.instrument) {
                                track.lastFn[(_c = ce.channel) !== null && _c !== void 0 ? _c : 0].push(track.instrument(note, system.currentTime(), track.pan));
                            }
                        }
                    }
                    if (ce.subtype === "EndOfTrack") {
                        stopAllSounds(track);
                    }
                }
            }
        }
        return [tempo, sig];
    };
    createPlayHeadClock(system, tempo, sig, tickCallBack);
}
